/*
 * mac_interface.c
 */

#include <QDOffscreen.h>
#include <Quickdraw.h>

#include "memory.h"
#include "debugger.h"
#include "sim6502.h"
#include "game.h"


WindowPtr	gWindow = NULL;
GWorldPtr	gScreenGWorld = NULL;

RGBColor	gColorValue [8][16];

RGBColor	black = {0,0,0};
RGBColor	white = {0xFFFF,0xFFFF,0xFFFF};
RGBColor	green = {0,0xFFFF,0};

GWorldPtr	currentPort;
GDHandle	currentDevice;


void	MacAllocateGraphicStructures 	(void);
void	MacDisposeGraphicStructures		(void);
void	MacInitColors 					(void);
short 	MacIsKeyPressed					(unsigned short k );


void InitApplication(void)
{
	short	err;
	
	InitGraf(&qd.thePort);
	InitFonts();
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(nil);
	InitCursor();
}


void draw_line (int x1, int y1, int x2, int y2, int color, int z)
{
	unsigned long 	pixel;
	RGBColor	pixelColor;
	RGBColor	savedFgColor;
	
	//	Sanity Check, be sure we have been initialized
	//
	if (!gScreenGWorld)
	{
		exit();
	}

	if (z != 0)
	{
		if ((x1 == x2) && (y1 == y2))
			x2++;

    	pixelColor = gColorValue [color & 0x07] [z];
		
		switch(game){
		 case LUNAR_LANDER:
		 case ASTEROIDS:
		 case ASTEROIDS_DX:
		 case RED_BARON:
		 case TEMPEST:
		  RGBForeColor(&white);
		  break;
		 
		 case BATTLEZONE:
		  RGBForeColor(&green);
		  break;

		 default:
		  RGBForeColor(&pixelColor);
		  break;
		}

		MoveTo(x1 >> 1 , (y1 >>1) +20);
		LineTo(x2>>1, (y2>>1) + 20);
	}
}

/*
 * This is called at vg startup time
 * Set up to draw to offscreen bitmap
 * and clear the bitmap to black
 */
void open_page (int step)
{
	//	Sanity Check, be sure we have been initialized
	//
	if (!gScreenGWorld)
	{
		exit();
	}
	SetGWorld(gScreenGWorld,NULL);
	RGBForeColor(&black);
	PaintRect(&gScreenGWorld->portRect);	
}

/*
 * This is called when drawing is done
 * Push the offscreen bitmap out to the display
 * leaving the display pointing to the screen
 */

void close_page (void)
{
	//	Sanity Check, be sure we have been initialized
	//
	if (!gScreenGWorld)
	{
		exit();
	}
	RGBForeColor(&black);
	RGBBackColor(&white);
	CopyBits((BitMap 
*)(*(gScreenGWorld->portPixMap)),&((GrafPtr)gWindow)->portBits,&gScreenGWorld->po
rtRect,&gWindow->portRect,srcCopy,nil);
	SetGWorld(currentPort,currentDevice);	
}


void init_graphics (int argc, char *argv[], int p_smallwindow, 
		    int p_use_pixmap, int p_line_width, char *window_name)
{
	MacInitColors();

	if (!gWindow)
	{		
	 MacAllocateGraphicStructures ();

	}
}

void exitCleanup()
{
	MacDisposeGraphicStructures();
	ExitToShell();
}


void handle_input ()
{
  /* Check to see if Quit Key (Q) is held down */
  if (MacIsKeyPressed (0x0C))
    {
      MacDisposeGraphicStructures ();
      ExitToShell ();
    }

  start1 = MacIsKeyPressed (0x12);
  start2 = MacIsKeyPressed (0x13);

  switches [0].left   = MacIsKeyPressed (0x02);
  switches [0].right  = MacIsKeyPressed (0x03);
  switches [0].fire   = MacIsKeyPressed (0x28);
  switches [0].thrust = MacIsKeyPressed (0x26);
  switches [0].hyper  = MacIsKeyPressed (0x04);
  switches [0].shield = MacIsKeyPressed (0x31);

#if 0
  switches [0].abort  = MacIsKeyPressed (0x35);

  cslot_left          = MacIsKeyPressed (0x18);
  cslot_right         = MacIsKeyPressed (0x35);
  cslot_util          = MacIsKeyPressed (0x33);
#endif

  if (MacIsKeyPressed (0x35))
    {
      stepflag = 1;
      traceflag = 0;
    }
}


short MacIsKeyPressed (unsigned short k)
/* k = any keyboard scan code, 0-127 */
{
  unsigned char km[16];

  GetKeys ((unsigned long *) km);
  return ((km [k >> 3] >> (k & 7)) & 1);
}


void MacAllocateGraphicStructures (void)
{
	QDErr	err;

	gWindow = GetNewWindow (128, NULL, (WindowPtr)-1L);
	if (gWindow==NULL)
		ExitToShell();

	//Temporary hack, it may not want to be like this eventually..
	err = NewGWorld(&gScreenGWorld, 8, &gWindow->portRect, NULL, NULL, 0);
	
	if (err != noErr)
	{
		DisposeWindow(gWindow);
		ExitToShell();
	}

	RGBForeColor(&white);
	RGBBackColor(&black);
	GetGWorld(&currentPort,&currentDevice);
	SetGWorld(gScreenGWorld, NULL);
	// clear offscreen pixmap
	PaintRect(&gWindow->portRect);
	SetGWorld(currentPort,currentDevice);
	// clear contents of window
	CopyBits((BitMap 
*)(*(gScreenGWorld->portPixMap)),&((GrafPtr)gWindow)->portBits,&gWindow->portRect
,&gWindow->portRect,srcCopy,nil);
}


void MacDisposeGraphicStructures(void)
{
	//	Sanity Check
	//
	if (gScreenGWorld==NULL)
		return;
	
	UnlockPixels(gScreenGWorld->portPixMap);
	SetGWorld(currentPort,currentDevice);
	DisposeGWorld(gScreenGWorld);
	gScreenGWorld=NULL;
	DisposeWindow(gWindow);
}

void MacInitColors (void)
{
short	i;
short	z;
short	color;

   i = 15 * 7 - 1;
   for (z = 15; z >= 0; z--)
	for (color = 0; color < 8; color++)
	  if ((color == 0) || (z == 0))
	    gColorValue [color] [z] = black;
	  else
	    {
	      if (i >= 0)
		{
		gColorValue[color] [z].red=(color & 4) ? (0x8000 + (0x0888 * z)) : 0;
		gColorValue[color] [z].green=(color & 2) ? (0x8000 + (0x0888 * z)) : 0;
		gColorValue[color] [z].blue= (color & 1) ? (0x8000 + (0x0888 * z)) : 0;
		}
	     else
		gColorValue [color] [z] = gColorValue [color] [z+1];
	    }
}





void calculate_inst_per_sec (unsigned long icount)
{
static	short beenhere=0;
static	unsigned long	lastcount;
static	unsigned long	lasttime;
char	theSpeed[256];

return;

if (!beenhere)
{
lastcount=icount;
lasttime=TickCount();
beenhere=1;
}
else
{
double	speed;
char	cycles[256];
char	seconds[256];


if (TickCount()-lasttime < 60)
return;


speed=(icount-lastcount)/1000000;

//speed=((float)(60*(icount-lastcount)))/((float)(TickCount()-lasttime));

sprintf(cycles,"%d",icount-lastcount); 
c2pstr(cycles);
DebugStr((StringPtr)cycles);

sprintf(seconds,"%d",(TickCount()-lasttime)/60); 
c2pstr(seconds);
DebugStr((StringPtr)seconds);


lasttime=TickCount();
lastcount=icount;

sprintf(theSpeed,"%f",speed); 
c2pstr(theSpeed);
DebugStr((StringPtr)theSpeed);
}


}


void calculate_mhz(void)
{
static	lasttick=0;
char	theSpeed[256];
float	speed;
char	seconds[256];

return;

if (lasttick==0)
{
lasttick=TickCount();
return;
}

speed=(1000000.0/60.0)/(((float)(TickCount()-lasttick)));

sprintf(seconds,"%d",(TickCount()-lasttick)); 
c2pstr(seconds);
DebugStr((StringPtr)seconds);

lasttick=TickCount();
sprintf(theSpeed,"%f",speed); 
c2pstr(theSpeed);
DebugStr((StringPtr)theSpeed);
}

