/*
 * pr_interface.c: Sun pixrect display for Atari Vector game simulator
 *
 * Copyright 1993, 1996 Eric Smith
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: pr_interface.c 2 2003-08-20 01:51:05Z eric $
 */

#include <stdio.h>

#include <pixrect/pixrect_hs.h>

#include "display.h"
#include "memory.h"   /* for switches[], start1, and start2 */
#include "sim6502.h"  /* for stepflag and traceflag */
#include "game.h"     /* for define of LUNAR_LANDER */


Pixrect *pr;
Pixrect *pr1;
Pixrect *pr2;


int smallwindow;
int window_width, window_height;


int use_double_buffer = 0;
int current_page;

int attr_list_a [] = 
{ 
  PR_DBL_DISPLAY, PR_DBL_A,
  PR_DBL_WRITE,   PR_DBL_B,
  PR_DBL_READ,    PR_DBL_B,
  PR_DBL_DISPLAY_DONTBLOCK, PR_DBL_A,
  0, 0
};
int attr_list_b [] =
{ 
  PR_DBL_DISPLAY, PR_DBL_B,
  PR_DBL_WRITE,   PR_DBL_A,
  PR_DBL_READ,    PR_DBL_A,
  PR_DBL_DISPLAY_DONTBLOCK, PR_DBL_B,
  0, 0
};


void setup_keyboard (void)
{
#if 0
  int i;

  for (i = 0; i < MAX_KEY; i++)
    keymap [i] = NULL;

  keymap [XKeysymToKeycode (mydisplay, XK_1)] = & start1;
  keymap [XKeysymToKeycode (mydisplay, XK_2)] = & start2;

  keymap [XKeysymToKeycode (mydisplay, XK_0)] = & self_test;

  keymap [XKeysymToKeycode (mydisplay, XK_d)] = & switches [0].left;
  keymap [XKeysymToKeycode (mydisplay, XK_f)] = & switches [0].right;
  keymap [XKeysymToKeycode (mydisplay, XK_k)] = & switches [0].fire;
  keymap [XKeysymToKeycode (mydisplay, XK_j)] = & switches [0].thrust;
  keymap [XKeysymToKeycode (mydisplay, XK_h)] = & switches [0].hyper;
  keymap [XKeysymToKeycode (mydisplay, XK_space)] = & switches [0].shield;
  keymap [XKeysymToKeycode (mydisplay, XK_q)] = & switches [0].abort;

  escape_code = XKeysymToKeycode (mydisplay, XK_Escape);
#endif
}


void init_graphics (int argc, char *argv[], int p_smallwindow, 
		    int p_use_pixmap, int p_line_width, char *window_name)
{
  smallwindow = p_smallwindow;

  window_width = 1024 >> p_smallwindow;
  window_height = 768 >> p_smallwindow;

  pr = pr_open ("/dev/fb");
  if (! pr)
    {
      fprintf (stderr, "Failed to open pixrect for screen!\n");
      exit (2);
    }

  if (pr)
    fprintf (stderr, "screen size: %d x %d, %d bits\n", pr->pr_size.x, pr->pr_size.y, pr->pr_depth);

  if ((pr->pr_size.x < window_width) || (pr->pr_size.y < window_height))
    {
      fprintf (stderr, "Screen too small!\n");
      exit (2);
    }

  if ((pr->pr_size.x == window_width) && (pr->pr_size.y == window_height))
    pr1 = pr;
  else
    {
      /* create a secondary pixrect that is the size we want and centered */
      pr1 = pr_region (pr, 
		       (pr->pr_size.x - window_width) / 2,
		       (pr->pr_size.y - window_height) / 2,
		       window_width, window_height);
      if (! pr1)
	{
	  pr1 = pr;
	  fprintf (stderr, "Couldn't create secondary pixrect!\n");
	}
    }

  if (p_use_pixmap)
    {
      int foo = pr_dbl_get (pr, PR_DBL_AVAIL);
      printf ("double buffer value = %d\n", foo);
      if (foo == PR_DBL_EXISTS)
	{
	  pr2 = pr1;
	  fprintf (stderr, "Hardware double-buffering enabled!\n");
	  use_double_buffer = 1;
	}
      else
	{
	  pr2 = mem_create (window_width, window_height, pr1->pr_depth);
	  
	  if (! pr2)
	    {
	      pr2 = pr1;
	      fprintf (stderr, "Failed to open pixrect for off-screen bitmap!\n");
	    }
	}
    }
  else
    pr2 = pr1;

  if (use_double_buffer)
    {
      pr_dbl_set (pr, attr_list_b);
      pr_rop (pr, 0, 0, pr->pr_size.x, pr->pr_size.y, PIX_SET, NULL, 0, 0);
      pr_dbl_set (pr, attr_list_a);
      pr_rop (pr, 0, 0, pr->pr_size.x, pr->pr_size.y, PIX_SET, NULL, 0, 0);
      current_page = PR_DBL_B;
    }
  else
    {
      /* clear the screen */
      pr_rop (pr, 0, 0, pr->pr_size.x, pr->pr_size.y, PIX_SET, NULL, 0, 0);
    }

  setup_keyboard ();
}


void term_graphics (void)
{
  if (pr1 != pr2)
    pr_close (pr2);
  pr_close (pr1);
}


void handle_input (void)
{
}


void draw_line (int x1, int y1, int x2, int y2, int c, int z)
{
  if ((c != 0) && (z != 0))
    pr_vector (pr2, x1 >> smallwindow, y1 >> smallwindow, 
	       x2 >> smallwindow, y2 >> smallwindow, PIX_CLR, 0);
}


void open_page (int step)
{
  /* clear pr2 */
  pr_rop (pr2, 0, 0, window_width, window_height, PIX_SET, NULL, 0, 0);
}


void close_page (void)
{
  if (use_double_buffer)
    {
      if (current_page == PR_DBL_A)
	{
	  pr_dbl_set (pr, attr_list_b);
	  current_page = PR_DBL_B;
	}
      else
	{
	  pr_dbl_set (pr, attr_list_a);
	  current_page = PR_DBL_A;
	}
    }
  else if (pr1 != pr2)
    {
      /* copy pr2 to pr1 */
      pr_rop (pr1, 0, 0, window_width, window_height, PIX_SRC, pr2, 0, 0);
    }
}
